"""Patch objects."""
from circadapt.components import Component


class Patch2022(Component):
    """
    Patch2022 is based on Patch in Walmsley 2015.

    Parameters
    ==========
    Am_ref: float
        Reference wall area at :math:`l_{s} = l_{s,ref}`.
    V_wall: float
        Wall volume
    v_max: float
        Maximum shortening velocity
    l_se0: float
        lgth of the series elastic element, i.e. :math:`l_{s} -l_{si}` for
        which stress is zero.
    l_s0: float
        Reference sarcomere lgth for which at
        :math:`A_m (l_{s,ref}) = A_{m,ref}`.
    l_s0: float
        Zero passive stress lgth
    dl_s_pas: float
        -
    Sf_pas: float
        Linear ECM stress coefficient
    k1: float
        Nonlinear exponent ECM stress component
    tr: float
        Contraction time constant
    td: float
        Relaxation time constant
    time_act: float
        Relative contraction duration
    Sf_act: float
        Linear active stress component
    dt: float
        Activation delay relative to intrinsic activation
    C_rest: float
        Rest contractility
    l_si0: float
        Reference lgth for zero-active-stress
    LDAD: float
        lgth dependend activation duration
    ADO: float
        activation duration offset
    LDCC: float
        lgth dependend contractility coefficient
    Sf_pasMaxT: float
        Maximum ecm stress (adaptation sens variable)
    Sf_pasActT: float
        Active weighted passive stress (adaptation sens variable)
    FacSf_actT: float
        Active stress (adaptation sens variable)
    LsPasActT: float
        Weighted sarcomere lgth average (adaptation sens variable)
    adapt_gamma: bool
        Adaptation constant

    Signals
    =======
    Signals are arrays. Each point in the array represents a point in time with
    step-size controlled by the solver.

    l_s: array
        Sarcomere lgth
    l_si: array
        State variable: Intrinsic sarcomere lgth
    LsiDot: array
        State variable: Intrinsic sarcomere lgth time-derivative
    C: array
        State variable: contraction curve
    C_dot: array
        State variable: contraction time-derivative
    Am: array
        Patch mid-wall area
    Am0: array
        Patch mid-wall zero-stress area
    Ef: array
        Natural strain
    T: array
        Mid-wall tension
    dA_dT: array
        Area-tension derivative
    Sf: array
        Total fibre stress at mid-wall
    Sf_pasT: array
        Total passive stress at mid-wall
    SfEcm: array
        Total ECM stress at mid-wall
    dSf_dEf: array
        Total stiffness coefficient
    dSf_pas_dEf: array
        Total passive stiffness coefficient
    SfEcmMax: array
        Adaptation: Maximum ECM stress
    Sf_actMax: array
        Adaptation: maximum active stress
    Sf_pasAct: array
        Adaptation: active-weighted passive stress
    LsPasAct: array
        Adaptation: active-weigthed sarcomere lgth

    """

    parameters = [
        'Am_ref',
        'V_wall',
        'v_max',
        'l_se0',
        'l_s0',
        'l_s_ref',
        'dl_s_pas',
        'Sf_pas',
        'tr',
        'td',
        'time_act',
        'Sf_act',
        'k1',
        'dt',
        'C_rest',
        'l_si0',
        'LDAD',
        'ADO',
        'LDCC',
        'SfPasMaxT',
        'SfPasActT',
        'FacSfActT',
        'LsPasActT',
        'adapt_gamma',
        ]

    signals = [
        'l_s',
        'l_si',
        'l_si_dot',
        'C',
        'C_dot',
        'Am',
        'Am0',
        'Ef',
        'T',
        'dA_dT',
        'Sf',
        'Sf_pas_total',
        'Sf_ecm',
        'dSf_dEf',
        'dSf_pas_dEf',
        'SfEcmMax',
        'SfActMax',
        'SfPasAct',
        'LsPasAct'
        ]

    def __init__(self, model, objects=None, wall_locs=None):
        """
        Patch2022 is based on Patch in Walmsley 2015.

        Number of patches and their names may vary and are automatically
        detected from the model object.

        Parameters
        ----------
        model : circadapt.Circadapt
            Model object.
        objects : list, optional
            List of objects in c++ this object refers to. If not given,
            the list is build based on wall_locs.
        wall_locs : list, optional
            List of wall locations in c++. This list is used to generate the
            objects.

        """
        self._model = model
        self._wall_locs = wall_locs
        if objects is None:
            objects = self.get_all_patches()
        super().__init__(model, objects)

    def build(self):
        """
        (re)build patch locations based on wall locations.

        There is no compatibility with different types of patches used in the
        model. If you use different type of patches, please use the direct way
        of interacting with the c++ object, i.e. set() and get().
        """
        self.objects = self.get_all_patches()
        super().build()

    def get_all_patches(self):
        """Get all patches of the walls included in this pointer object."""

        # get all wall objects from all possible walls
        if self._wall_locs is None:
            _wall_locs = []
            for key in ['Wall2022', 'PreAfterloadExperiment']:
                if key in self._model.components:
                    _wall_locs += self._model[key].objects
        else:
            _wall_locs = self._wall_locs

        all_patches = []
        for i_wall, wall in enumerate(_wall_locs):
            n_patch = self._model.get(wall + '.n_patch', int)
            for i_patch in range(n_patch):
                patch_name = self._model.get(wall + '.subcomponents:' +
                                             str(i_patch), str)
                all_patches.append(wall+'.'+patch_name)
        return all_patches
