"""
Benchmark models
"""

import circadapt
import circadapt.plot
from circadapt.model import Model
from circadapt.adapt import ModelAdapt
import matplotlib.pyplot as plt
import numpy as np


class Patch2022(Model):
    def __init__(self,
                 solver: str = None,
                 path_to_circadapt: str = None,
                 model_state: dict = None,
                 ):
        if solver is None:
            solver = 'forward_euler'

        ModelAdapt.__init__(self)
        Model.__init__(self,
                       solver,
                       path_to_circadapt=path_to_circadapt,
                       model_state=model_state,
                       )

    def build(self):
        """
        Build pre-afterload model with Patch2022.

        This model is designed for illustrating the Patch2022 isolated from the
        circulation and for benchmarking the Patch2022 module.
        """
        self.add_component('PreAfterloadExperiment', 'W')
        self.add_component('Patch2022', 'P', 'W')

    def set_reference(self):
        self['Patch2022']['dt'] = 0.1
        self['Patch2022']['time_act'] = 0.3
        self['Patch2022']['Sf_act'] = 120e3
        self['Patch2022']['v_max'] = 7

        self['PreAfterloadExperiment']['Am_dead'] = 0
        self['PreAfterloadExperiment']['Am_ref_afterload'] = 1.3 * self['Patch2022']['Am_ref']
        self['PreAfterloadExperiment']['T_afterload'] = 50

        self.set('Model.W.P.l_si', 1.0, float)

        # update assert test settings
        self.assert_test_settings['l_si_t0'] = self.get('Model.W.P.l_si',
                                                        float)
    def assert_1_beat(self):
        """
        Test status after 1 beat, used in unit tests.

        This function can be tested after creating the model with the
        implemented reference and after 1 beat has been calculated. In case
        the initial model state deviates from the implemented reference, test
        settings might be changed accordingly.

        Test for:
            l_si == l_si_init  on t=t0
            l_se_norm == 1     on t=tend
        """
        l_si_t0 = self['Patch2022']['l_si'][0, 0]
        assert (l_si_t0 == self.assert_test_settings['l_si_t0'])

        l_si_tend = self['Patch2022']['l_si'][-1, 0]
        l_s_tend = self['Patch2022']['l_s'][-1, 0]
        l_se = self['Patch2022']['l_se0'][0]
        l_se_norm = (l_s_tend-l_si_tend) / l_se
        assert np.abs(l_se_norm-1) < 1e-9

    def plot(self, fig=None):
        if fig is None:
            fig = plt.figure(1, clear=True, figsize=(8, 6))
        m = 2
        n = 2

        t = self.get('Solver.t') * 1e3

        ax1 = fig.add_subplot(m, n, 1)
        ax1.plot(t, self['Patch2022']['l_s'])
        ax1.plot(t, self['Patch2022']['l_si'])
        ax1.legend(['l_s', 'l_si'])
        ax1.set_ylabel('Sarcomere length [$\mu m$]')

        ax2 = fig.add_subplot(m, n, 2)
        ax2.plot(t, self['Patch2022']['C'], label='C')
        ax2.set_ylabel('C-curve [-]')

        ax3 = fig.add_subplot(m, n, 3)
        ax3.plot(t, self['Patch2022']['Sf']*1e-3, label='Sf')
        ax3.plot(t, self['Patch2022']['Sf_ecm']*1e-3, label='SfEcm')
        ax3.set_ylabel('Stress [kPa]')
        ax3.legend()

        ax4 = fig.add_subplot(m, n, 4)
        ax4.plot(t, self['Patch2022']['T'], label='Tension Wall')
        ax4.set_ylabel('T [$N/m$]')

        # Plot design
        for ax in [ax1, ax2, ax3, ax4]:
            ax.spines[['right', 'top']].set_visible(False)
        ax3.set_xlabel('Time [ms]')
        ax4.set_xlabel('Time [ms]')

        plt.tight_layout()

